/***********************************
*   EJERCICIO 8-3
*   Estimadores de estado
************************************/

xdel(winsid());
clear all;
clc;

// Parámetros del sistema //
// Motor de CC
Ra = 2.4;           // Resistencia de armadura
La = 0.04;          // Inductancia de armadura
Kt = 0.86;          // Constante de par
Kw = 1.07;          // Constante de velocidad
A0 = 10;            // Ganancia amplificador
// Carga Mecánica
B1 = 0.04;          // Rozamiento eje motor
B2 = 2;             // Rozamiento lámina
J = 0.025;          // Momento de inercia eje motor
K = 500;            // Constante elástica resorte
r1 = 0.1;           // Radio rodillo
// Sensor
Kp = 1;             // Ganancia sensor de posición

// Modelo continuo //
// Matrices del modelo de estado
A = [-Ra/La -Kw/La 0; Kt/J -B1/J -K*r1/(2*J); 0 r1/2 -K/(4*B2)];
B = [A0/La; 0; 0];
C = [0 0 Kp];
D = 0;
// Definición del modelo de estado
P1 = syslin("c",A,B,C,D)
// Función transferencia
G1 = ss2tf(P1)
// Observabilidad //
V = obsv_mat(P1)
rank(V)
// Simulación //
// Modelo con salidas de todas las variables de estado para graficar su respuesta
// a un escalon en la entrada
P1_sim = P1;
P1_sim.c = [1 0 0; 0 1 0; C];
P1_sim.d = [0;0;0];
T = 0:0.005:0.5;            // Eje temporal
Y=csim('step',T,P1_sim);    // Respuesta al escalón
// Gráfico:
subplot(3,1,1)
plot2d(T,Y(1,:),color("red"));
xgrid(color("gray"))
title('Corriente de armadura')
ylabel("[A]");
subplot(3,1,2)
plot2d(T,Y(2,:),color("scilabgreen4"));
xgrid(color("gray"))
title('Velocidad angular del rotor')
ylabel("[rad/s]]");
subplot(3,1,3)
plot2d(T,1000*Y(3,:),color("blue"));
xgrid(color("gray"))
title('Tensión del sensor de posición del resorte')
ylabel("[mV]");

// Modelo discreto //
Ts = 0.01;          // Periodo de muestreo
P2 = dscr(P1,Ts);   // Modelo discreto con periodo de muestreo Ts
// Matrices del modelo discreto:
Ad=P2.a;
Bd=P2.b;
Cd=P2.c;
Dd=P2.d;
// Autovalores del modelo discreto
autov=spec(Ad);
// Función transferencia
G2 = ss2tf(P2);
 // Observabilidad //
V2 = obsv_mat(P2)
rank(V2)
// Simulación //
// Modelo con salidas de todas las variables de estado para graficar su respuesta
// a un escalon en la entrada
P2_sim = P2;
P2_sim.c = [1 0 0; 0 1 0; Cd];
P2_sim.d = [0;0;0];
T_dis = 0:Ts:0.5;                   // Vector tiempo discreto
dstep = ones(1,length(T_dis));      // Escalón discreto
Y2 = dsimul(P2_sim,dstep);          // Respuesta al escalón
// Gráfico:
figure(1)
subplot(3,1,1)
plot2d2(T_dis,Y2(1,:),color("red"));
xgrid(color("gray"))
title('Corriente de armadura')
ylabel("[A]");
subplot(3,1,2)
plot2d2(T_dis,Y2(2,:),color("scilabgreen4"));
xgrid(color("gray"))
title('Velocidad angular del rotor')
ylabel("[rad/s]]");
subplot(3,1,3)
plot2d2(T_dis,1000*Y2(3,:),color("blue"));
xgrid(color("gray"))
title('Tensión del sensor de posición del resorte')
ylabel("[mV]");

// Modelo Canónico Observable (MCO) //
clc;
den = coeff(G2.den)                         // Denominador de la FT discreta
num = coeff(G2.num)                         // Numerador de la FT discreta
// Matrices del MCO:
Ac = [0,0,-den(1);1,0,-den(2);0,1,-den(3)];
Cc = [0,0,1];
Bc = [num(1);num(2);num(3)];
Dc = 0;
// Definición del MCO:
str1 = sprintf(' Modelo Canónico Observable ');
disp(str1);
P3 = syslin("d",Ac,Bc,Cc,Dc)
// Observabilidad
Vco = obsv_mat(P3)
det(Vco)
// Matriz de transformación:
str1 = sprintf('Matriz de Transformación al MCO');
disp(str1);
Pco = V2^-1*Vco
Pco_1 = Pco^-1
// Verifico la matriz de transformación:
// 1. Comparando matrices A, B y C
Aco = Pco_1*Ad*Pco
Bco = Pco_1*Bd
Cco = Cd*Pco
// 2. Transformando al MCO con función embebida:
P4 = ss2ss(P2,Pco) 

// Diseño del estimador //
clc;
// Autovalores deseados:
z1 = 0;
z2 = 0;
z3 = 0;
autov_d = [z1,z2,z3];
// Vector de realimentación H:
Pcr = poly(autov_d,"z","roots");    // Polinomio característico deseado
Pcm = poly(autov,"z","roots");      // Polinomio característico de la planta
// Cálculo de los elementos del vector H
Pcd = Pcr-Pcm;
Hco = (coeff(Pcd))'                 // Vector de realimentación H del MCO
Hco = (ppol(Aco',Cco',autov_d))'    // Verifico usando función embebida
H = Pco*Hco                         // Vector de realimentación H del modelo original
H = (ppol(Ad',Cd',autov_d))'        // Verifico usando funcion embebida
// Calculo de H con mínima norma
Nmin=1e23;
zmin=[0 0 0];
for z1=0.5:0.02:1
    for z2=0.5:0.02:1
        for z3=0.5:0.02:1
            z=[z1 z2 z3];
            H1 = (ppol(Ad',Cd',z))'
            N=norm(H1,2);
            if N < Nmin
                Nmin = N;
                zmin = z;
                Hmin = H1;
            end 
        end
    end
end
zmin
Hmin
Nmin

//***************************************************//
// Descomentar si se quiere usar el H de norma mínima
H = Hmin
//***************************************************//

// Matrices
Ae = P2.a-H*P2.c
Be = [P2.b,H]
Ce = eye(3,3)
// Autovalores
autov_e = spec(Ae)

// Modelo del error de estimación
xdel(winsid());
clc;
Xe0 = [1;1;0.001]                   // Condiciones iniciales del error
Ve  =zeros(3,1);                    // Matriz B o D del modelo del error
Pe = syslin("d",Ae,Ve,Ce,Ve,Xe0)    // Definición del modelo del error
// Simulación
u = zeros(1,length(T_dis));         // Entrada nula al modelo del error
Y3 = dsimul(Pe,u);                  // Respuesta a las condiciones iniciales
// Gráfico:
subplot(3,1,1)
plot2d2(T_dis,Y3(1,:),color("red"));
xgrid(color("gray"))
title('Error estimación corriente de armadura')
ylabel("[A]");
subplot(3,1,2)
plot2d2(T_dis,Y3(2,:),color("scilabgreen4"));
xgrid(color("gray"))
title('Error estimación velocidad angular del rotor')
ylabel("[rad/s]]");
subplot(3,1,3)
plot2d2(T_dis,1000*Y3(3,:),color("blue"));
xgrid(color("gray"))
title('Error estimación tensión del sensor de posición del resorte')
ylabel("[mV]");

// Modelo completo (Modelo original + Estimador)
// Matrices del modelo de estado:
//*******************************
At = [Ad,zeros(3,3);H*Cd,Ae]
Bt = [Bd;Bd]
Ct = eye(6,6)
Dt = zeros(6,1)
//*******************************
Xt = [Xe0;zeros(3,1)]           // Condiciones iniciales del modelo completo
Pt = syslin("d",At,Bt,Ct,Dt,Xt) // Definición del modelo completo
// Simulación
Y4 = dsimul(Pt,dstep);          // Respuesta al escalón
// Gráfico
subplot(3,1,1)
plot2d2(T_dis,Y4(1,:),color("red"));
plot2d2(T_dis,Y4(4,:),color("blue"));
xgrid(color("gray"))
title('Respuesta de las variables de estado')
ylabel("Corriente");
subplot(3,1,2)
plot2d2(T_dis,Y4(2,:),color("red"));
plot2d2(T_dis,Y4(5,:),color("blue"));
xgrid(color("gray"))
ylabel("Velocidad");
subplot(3,1,3)
plot2d2(T_dis,Y4(3,:),color("red"));
plot2d2(T_dis,Y4(6,:),color("blue"))
xgrid(color("gray"))
ylabel("Posición");






